﻿using System.Linq;
using System.Collections.Generic;
using Ato.EN.IntegrationServices.CodeGenerationIEE;
using DataContracts;
using Au.Gov.Sbr.Xbrl.Metadata;
using Au.Gov.Sbr.Xbrl.Document;

namespace Ato.CD.Inbound.IEE201802
{
    /// <summary>
    /// Subclass of Consumer class which implements rules that require access to unconsumed payload data to be implemented correctly
    /// </summary>
    class IEE2018ConsumerValidateUnsupportedRules : IEE2018Consumer
    {
        /// <summary>
        /// Validates rules that can not be generated and must run at a lower level.
        /// Overrides empty abstract method defined in IEE2018Consumer.
        /// </summary>
        /// <param name="sbrDocument">The SBR document.</param>
        /// <param name="populatedReportElementsMap">The populated report elements map can be used to find data elements and tuples.</param>
        /// <param name="contextArray">The context array can be used to find context values.</param>
        /// <param name="errors">The errors collection is to be populated with validation results.</param>
        protected override void ValidateUnsupportedRules(SBRDocument sbrDocument, IDictionary<string, SBRElement> populatedReportElementsMap, SBRContext[] contextArray, List<ProcessMessageDocument> errors)
        {
            base.ValidateUnsupportedRules(sbrDocument, populatedReportElementsMap, contextArray, errors);

            ProcessMessageDocument processMessage;

            /// <summary>
            /// Current element being extracted
            /// </summary>
            SBRElement currentValue;
            /// <summary>
            /// Element value retrieved 
            /// </summary>
            string parameterOutput;

            #region Populate sbrRPContexts
            IEnumerable<SBRContext> sbrRPContexts = null;
            SBRContext sbrRPContext = null;

            sbrRPContexts = contextArray.Where(
                c => c.Period != null && c.Period.PeriodType == SBRPeriod_PeriodType.DURATION &&
                c.Segment != null && c.Segment.ExplicitOrTypedDimensionDomains.Count == 1 &&
                (c.Segment.ExplicitOrTypedDimensionDomains.Any(dd => dd.DimensionType == SBRExplicitOrTypedDimensionDomain_Type.EXPLICIT && dd.Dimension.EndsWith("}ReportPartyTypeDimension") && ((SBRExplicitDimensionDomain)dd).Domain.EndsWith("}ReportingParty"))));

            if (sbrRPContexts != null && sbrRPContexts.Count() > 0)
            {
                sbrRPContext = sbrRPContexts.First();
            }
            #endregion

            #region Populate root-level SBRElement collections   
            #region Populate sbrPeriodDetailsCollection
            IEnumerable<SBRElement> sbrPeriodDetailsCollection;
            if (populatedReportElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}PeriodDetails", out currentValue))
                sbrPeriodDetailsCollection = currentValue.Occurrences;
            else
                sbrPeriodDetailsCollection = null;
            #endregion

            #region Populate sbrDeclarationCollection
            IEnumerable<SBRElement> sbrDeclarationCollection;
            if (populatedReportElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}Declaration", out currentValue))
                sbrDeclarationCollection = currentValue.Occurrences;
            else
                sbrDeclarationCollection = null;
            #endregion

            #region Populate sbrTrusteeOrPartnerCollection
            IEnumerable<SBRElement> sbrTrusteeOrPartnerCollection;
            if (populatedReportElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}TrusteeOrPartner", out currentValue))
                sbrTrusteeOrPartnerCollection = currentValue.Occurrences;
            else
                sbrTrusteeOrPartnerCollection = null;
            #endregion

            #region Populate sbrFamilyTrustCollection
            IEnumerable<SBRElement> sbrFamilyTrustCollection;
            if (populatedReportElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}FamilyTrust", out currentValue))
                sbrFamilyTrustCollection = currentValue.Occurrences;
            else
                sbrFamilyTrustCollection = null;
            #endregion

            #region Populate sbrSpecifiedIndividualCollection
            IEnumerable<SBRElement> sbrSpecifiedIndividualCollection;
            if (populatedReportElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}SpecifiedIndividual", out currentValue))
                sbrSpecifiedIndividualCollection = currentValue.Occurrences;
            else
                sbrSpecifiedIndividualCollection = null;
            #endregion

            #region Populate sbrDeclarationOfTrusteeOrPartnerCollection
            IEnumerable<SBRElement> sbrDeclarationOfTrusteeOrPartnerCollection;
            if (populatedReportElementsMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}DeclarationOfTrusteeOrPartner", out currentValue))
                sbrDeclarationOfTrusteeOrPartnerCollection = currentValue.Occurrences;
            else
                sbrDeclarationOfTrusteeOrPartnerCollection = null;
            #endregion
            #endregion

            #region VR.ATO.IEE.410116
            #region VR.ATO.IEE.410116 - manual implementation for instances against IEE97, IEE98, IEE99

            /*  VR.ATO.GEN.410116
                Period code must be 'ControlOutsideAustraliaPartPeriod', 'NonResidentPartPeriod' or 'SubstitutedAccountingPeriod'

                Legacy Rule Format:
                    IF (pyin.xx.xx:Period.Type.Code <> NULLORBLANK) AND 
                        (pyin.xx.xx:Period.Type.Code <> SET("ControlOutsideAustraliaPartPeriod","NonResidentPartPeriod","SubstitutedAccountingPeriod"))
                       RETURN VALIDATION MESSAGE
                    ENDIF

                Technical Business Rule Format:
                    (^IEE97 <> BLANK) AND     (NotInSet(^IEE97, '"ControlOutsideAustraliaPartPeriod","NonResidentPartPeriod","SubstitutedAccountingPeriod"'))

                Data Elements:

                RP:^IEE97 = tns:Period.Type.Code WHERE [Period.Type.Code='ControlOutsideAustraliaPartPeriod']
*/


            IEnumerable<SBRElement> VRATOIEE410116_IEE97_Violators = null;
            if (sbrRPContexts != null && sbrPeriodDetailsCollection != null && sbrPeriodDetailsCollection.Count() > 0)
            {
                VRATOIEE410116_IEE97_Violators = sbrPeriodDetailsCollection.Where(
                    ce => ce.ChildElements != null &&
                    ce.ChildElements.Any(te => te.Name == "Period.Type.Code" && te.Value != "ControlOutsideAustraliaPartPeriod" && te.Value != "NonResidentPartPeriod" && te.Value != "SubstitutedAccountingPeriod" &&
                    sbrRPContexts.Contains(te.Context)));
            }

            if (VRATOIEE410116_IEE97_Violators != null && VRATOIEE410116_IEE97_Violators.Count() > 0)
            {
                foreach (SBRElement violator in VRATOIEE410116_IEE97_Violators)
                {
                    processMessage = new ProcessMessageDocument()
                    {
                        Code = "CMN.ATO.GEN.410116",
                        Severity = ProcessMessageSeverity.Error,
                        Description = @"Period code must be a valid code",
                        Location = DeriveSbrElementLocation(violator, addOccurrenceIndex: true) + "/tns:Period.Type.Code",
                        LongDescription = @"Period code must be 'ControlOutsideAustraliaPartPeriod', 'NonResidentPartPeriod' or 'SubstitutedAccountingPeriod'",
                        Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410116" } },
                    };


                    IDictionary<string, SBRElement> VRATOIEE410116_IEE97_Violator_TupleElementMap = violator.PopulatedChildElementsMap ?? new Dictionary<string, SBRElement>();

                    parameterOutput = "";

                    if (VRATOIEE410116_IEE97_Violator_TupleElementMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}Period.Type.Code", out currentValue))
                    {
                        if (!currentValue.IsNil) parameterOutput = currentValue.Value;
                    }

                    processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "IEE97", Value = parameterOutput });

                    errors.Add(processMessage);
                }
            }

            #endregion

            #region VR.ATO.IEE.410116 - manual implementation for instance against IEE102

            /*  VR.ATO.GEN.410116
                        Period code must be 'ControlOutsideAustraliaPartPeriod', 'NonResidentPartPeriod' or 'SubstitutedAccountingPeriod'

                    Legacy Rule Format:
                        IF (pyin.xx.xx:Period.Type.Code <> NULLORBLANK) AND 
                            (pyin.xx.xx:Period.Type.Code <> SET("ControlOutsideAustraliaPartPeriod","NonResidentPartPeriod","SubstitutedAccountingPeriod"))
                           RETURN VALIDATION MESSAGE
                        ENDIF

                Technical Business Rule Format:
                        (^IEE102 <> BLANK) AND     (NotInSet(^IEE102, '"ControlOutsideAustraliaPartPeriod","NonResidentPartPeriod","SubstitutedAccountingPeriod"'))

            Data Elements:

                        RP:^IEE102 = tns:Period.Type.Code WHERE [Period.Type.Code='NonResidentPartPeriod']
                        */
            if (sbrTrusteeOrPartnerCollection != null)
            {
                foreach (SBRElement sbrTrusteeOrPartner in sbrTrusteeOrPartnerCollection)
                {

                    IDictionary<string, SBRElement> sbrTrusteeOrPartnerTupleElementMap = sbrTrusteeOrPartner.PopulatedChildElementsMap ?? new Dictionary<string, SBRElement>();

                    IEnumerable<SBRElement> sbrTrusteeOrPartner_PeriodDetailsCollection;
                    if (sbrTrusteeOrPartnerTupleElementMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}PeriodDetails", out currentValue))
                        sbrTrusteeOrPartner_PeriodDetailsCollection = currentValue.Occurrences;
                    else
                        sbrTrusteeOrPartner_PeriodDetailsCollection = null;
                    if (sbrRPContexts != null && sbrTrusteeOrPartner_PeriodDetailsCollection != null && sbrTrusteeOrPartner_PeriodDetailsCollection.Count() > 0)
                    {
                        IEnumerable<SBRElement> VRATOIEE410116_IEE102_CurrentParentTupleViolators = sbrTrusteeOrPartner_PeriodDetailsCollection.Where(
                            ce => ce.ChildElements != null &&
                            ce.ChildElements.Any(te => te.Name == "Period.Type.Code" && te.Value != "ControlOutsideAustraliaPartPeriod" && te.Value != "NonResidentPartPeriod" && te.Value != "SubstitutedAccountingPeriod" &&
                            sbrRPContexts.Contains(te.Context)));
                        if (VRATOIEE410116_IEE102_CurrentParentTupleViolators != null && VRATOIEE410116_IEE102_CurrentParentTupleViolators.Count() > 0)
                        {
                            foreach (SBRElement violator in VRATOIEE410116_IEE102_CurrentParentTupleViolators)
                            {
                                processMessage = new ProcessMessageDocument()
                                {
                                    Code = "CMN.ATO.GEN.410116",
                                    Severity = ProcessMessageSeverity.Error,
                                    Description = @"Period code must be a valid code",
                                    Location = DeriveSbrElementLocation(violator, addOccurrenceIndex: true) + "/tns:Period.Type.Code",
                                    LongDescription = @"Period code must be 'ControlOutsideAustraliaPartPeriod', 'NonResidentPartPeriod' or 'SubstitutedAccountingPeriod'",
                                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410116" } },
                                };


                                IDictionary<string, SBRElement> VRATOIEE410116_IEE102_Violator_TupleElementMap = violator.PopulatedChildElementsMap ?? new Dictionary<string, SBRElement>();

                                parameterOutput = "";

                                if (VRATOIEE410116_IEE102_Violator_TupleElementMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}Period.Type.Code", out currentValue))
                                {
                                    if (!currentValue.IsNil) parameterOutput = currentValue.Value;
                                }

                                processMessage.Parameters.Add(new ProcessMessageParameter
                                { Name = "IEE102", Value = parameterOutput });

                                errors.Add(processMessage);
                            }
                        }


                    }
                }
            }

            #endregion

            #region VR.ATO.IEE.410116 - manual implementation for instance against IEE103

            /*  VR.ATO.GEN.410116
                        Period code must be 'ControlOutsideAustraliaPartPeriod', 'NonResidentPartPeriod' or 'SubstitutedAccountingPeriod'

                    Legacy Rule Format:
                        IF (pyin.xx.xx:Period.Type.Code <> NULLORBLANK) AND 
                            (pyin.xx.xx:Period.Type.Code <> SET("ControlOutsideAustraliaPartPeriod","NonResidentPartPeriod","SubstitutedAccountingPeriod"))
                           RETURN VALIDATION MESSAGE
                        ENDIF

                Technical Business Rule Format:
                        (^IEE103 <> BLANK) AND     (NotInSet(^IEE103, '"ControlOutsideAustraliaPartPeriod","NonResidentPartPeriod","SubstitutedAccountingPeriod"'))

            Data Elements:

                        RP:^IEE103 = tns:Period.Type.Code WHERE [Period.Type.Code='SubstitutedAccountingPeriod']
                        */
            if (sbrFamilyTrustCollection != null)
            {
                foreach (SBRElement sbrFamilyTrust in sbrFamilyTrustCollection)
                {

                    IDictionary<string, SBRElement> sbrFamilyTrustTupleElementMap = sbrFamilyTrust.PopulatedChildElementsMap ?? new Dictionary<string, SBRElement>();

                    IEnumerable<SBRElement> sbrFamilyTrust_PeriodDetailsCollection;
                    if (sbrFamilyTrustTupleElementMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}PeriodDetails", out currentValue))
                        sbrFamilyTrust_PeriodDetailsCollection = currentValue.Occurrences;
                    else
                        sbrFamilyTrust_PeriodDetailsCollection = null;
                    if (sbrRPContexts != null && sbrFamilyTrust_PeriodDetailsCollection != null && sbrFamilyTrust_PeriodDetailsCollection.Count() > 0)
                    {
                        IEnumerable<SBRElement> VRATOIEE410116_IEE103_CurrentParentTupleViolators = sbrFamilyTrust_PeriodDetailsCollection.Where(
                            ce => ce.ChildElements != null &&
                            ce.ChildElements.Any(te => te.Name == "Period.Type.Code" && te.Value != "ControlOutsideAustraliaPartPeriod" && te.Value != "NonResidentPartPeriod" && te.Value != "SubstitutedAccountingPeriod" &&
                            sbrRPContexts.Contains(te.Context)));
                        if (VRATOIEE410116_IEE103_CurrentParentTupleViolators != null && VRATOIEE410116_IEE103_CurrentParentTupleViolators.Count() > 0)
                        {
                            foreach (SBRElement violator in VRATOIEE410116_IEE103_CurrentParentTupleViolators)
                            {
                                processMessage = new ProcessMessageDocument()
                                {
                                    Code = "CMN.ATO.GEN.410116",
                                    Severity = ProcessMessageSeverity.Error,
                                    Description = @"Period code must be a valid code",
                                    Location = DeriveSbrElementLocation(violator, addOccurrenceIndex: true) + "/tns:Period.Type.Code",
                                    LongDescription = @"Period code must be 'ControlOutsideAustraliaPartPeriod', 'NonResidentPartPeriod' or 'SubstitutedAccountingPeriod'",
                                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.GEN.410116" } },
                                };


                                IDictionary<string, SBRElement> VRATOIEE410116_IEE103_Violator_TupleElementMap = violator.PopulatedChildElementsMap ?? new Dictionary<string, SBRElement>();

                                parameterOutput = "";

                                if (VRATOIEE410116_IEE103_Violator_TupleElementMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}Period.Type.Code", out currentValue))
                                {
                                    if (!currentValue.IsNil) parameterOutput = currentValue.Value;
                                }


                                processMessage.Parameters.Add(new ProcessMessageParameter
                                { Name = "IEE103", Value = parameterOutput });

                                errors.Add(processMessage);
                            }
                        }
                    }
                }
            }

            #endregion
            #endregion

            #region VR.ATO.IEE.410139
            #region VR.ATO.IEE.410139 - manual implementation for instance against IEE130

            /*  VR.ATO.IEE.410139
                        Address Currency Code must be 'C'

                    Legacy Rule Format:
                        IN TUPLE((iee.0003.lodge.request.xx.xx:TrusteeOrPartner) OR (iee.0003.lodge.request.xx.xx:FamilyTrust) OR (iee.0003.lodge.request.xx.xx:SpecifiedIndividual)) 
                        IN TUPLE(address2.xx.xx:AddressDetails) 
                        IF (pyde.xx.xx:AddressDetails.Currency.Code <> NULLORBLANK) AND (pyde.xx.xx:AddressDetails.Currency.Code <> "C")
                            RETURN VALIDATION MESSAGE 
                        ENDIF

                Technical Business Rule Format:
                        (^IEE130 <> BLANK) AND (^IEE130 <> 'C')

            Data Elements:

                        RP:^IEE130 = tns:AddressDetails.Currency.Code WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
                        */


            if (sbrTrusteeOrPartnerCollection != null)
            {
                foreach (SBRElement sbrTrusteeOrPartner in sbrTrusteeOrPartnerCollection)
                {

                    IDictionary<string, SBRElement> sbrTrusteeOrPartnerTupleElementMap = sbrTrusteeOrPartner.PopulatedChildElementsMap ?? new Dictionary<string, SBRElement>();

                    IEnumerable<SBRElement> sbrTrusteeOrPartner_AddressDetailsCollection;
                    if (sbrTrusteeOrPartnerTupleElementMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}AddressDetails", out currentValue))
                        sbrTrusteeOrPartner_AddressDetailsCollection = currentValue.Occurrences;
                    else
                        sbrTrusteeOrPartner_AddressDetailsCollection = null;
                    if (sbrRPContexts != null && sbrTrusteeOrPartner_AddressDetailsCollection != null && sbrTrusteeOrPartner_AddressDetailsCollection.Count() > 0)
                    {
                        IEnumerable<SBRElement> VRATOIEE410139_IEE130_CurrentParentTupleViolators = sbrTrusteeOrPartner_AddressDetailsCollection.Where(
                            ce => ce.ChildElements != null &&
                            ce.ChildElements.Any(te => te.Name == "AddressDetails.Currency.Code" && !string.IsNullOrWhiteSpace(te.Value) && te.Value != "C" &&
                            sbrRPContexts.Contains(te.Context)));
                        if (VRATOIEE410139_IEE130_CurrentParentTupleViolators != null && VRATOIEE410139_IEE130_CurrentParentTupleViolators.Count() > 0)
                        {
                            foreach (SBRElement violator in VRATOIEE410139_IEE130_CurrentParentTupleViolators)
                            {
                                processMessage = new ProcessMessageDocument()
                                {
                                    Code = "CMN.ATO.IEE.410139",
                                    Severity = ProcessMessageSeverity.Error,
                                    Description = @"Address Currency Code must be ""C""",
                                    Location = DeriveSbrElementLocation(violator, addOccurrenceIndex: true) + "/tns:AddressDetails.Currency.Code",
                                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.IEE.410139" } },
                                };


                                IDictionary<string, SBRElement> VRATOIEE410139_IEE130_Violator_TupleElementMap = violator.PopulatedChildElementsMap ?? new Dictionary<string, SBRElement>();

                                parameterOutput = "";

                                if (VRATOIEE410139_IEE130_Violator_TupleElementMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}AddressDetails.Currency.Code", out currentValue))
                                {
                                    if (!currentValue.IsNil) parameterOutput = currentValue.Value;
                                }

                                processMessage.Parameters.Add(new ProcessMessageParameter
                                { Name = "IEE130", Value = parameterOutput });

                                errors.Add(processMessage);
                            }
                        }
                    }
                }
            }
            #endregion

            #region VR.ATO.IEE.410139 - manual implementation for instance against IEE153

            /*  VR.ATO.IEE.410139
                        Address Currency Code must be 'C'

                    Legacy Rule Format:
                        IN TUPLE((iee.0003.lodge.request.xx.xx:TrusteeOrPartner) OR (iee.0003.lodge.request.xx.xx:FamilyTrust) OR (iee.0003.lodge.request.xx.xx:SpecifiedIndividual)) 
                        IN TUPLE(address2.xx.xx:AddressDetails) 
                        IF (pyde.xx.xx:AddressDetails.Currency.Code <> NULLORBLANK) AND (pyde.xx.xx:AddressDetails.Currency.Code <> "C")
                            RETURN VALIDATION MESSAGE 
                        ENDIF

                Technical Business Rule Format:
                        (^IEE130 <> BLANK) AND (^IEE130 <> 'C')

            Data Elements:

                        RP:^IEE130 = tns:AddressDetails.Currency.Code WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
                        */


            if (sbrFamilyTrustCollection != null)
            {
                foreach (SBRElement sbrFamilyTrust in sbrFamilyTrustCollection)
                {
                    IDictionary<string, SBRElement> sbrFamilyTrustTupleElementMap = sbrFamilyTrust.PopulatedChildElementsMap ?? new Dictionary<string, SBRElement>();

                    IEnumerable<SBRElement> sbrFamilyTrust_AddressDetailsCollection;
                    if (sbrFamilyTrustTupleElementMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}AddressDetails", out currentValue))
                        sbrFamilyTrust_AddressDetailsCollection = currentValue.Occurrences;
                    else
                        sbrFamilyTrust_AddressDetailsCollection = null;
                    if (sbrRPContexts != null && sbrFamilyTrust_AddressDetailsCollection != null && sbrFamilyTrust_AddressDetailsCollection.Count() > 0)
                    {
                        IEnumerable<SBRElement> VRATOIEE410139_IEE153_CurrentParentTupleViolators = sbrFamilyTrust_AddressDetailsCollection.Where(
                            ce => ce.ChildElements != null &&
                            ce.ChildElements.Any(te => te.Name == "AddressDetails.Currency.Code" && !string.IsNullOrWhiteSpace(te.Value) && te.Value != "C" &&
                            sbrRPContexts.Contains(te.Context)));
                        if (VRATOIEE410139_IEE153_CurrentParentTupleViolators != null && VRATOIEE410139_IEE153_CurrentParentTupleViolators.Count() > 0)
                        {
                            foreach (SBRElement violator in VRATOIEE410139_IEE153_CurrentParentTupleViolators)
                            {
                                processMessage = new ProcessMessageDocument()
                                {
                                    Code = "CMN.ATO.IEE.410139",
                                    Severity = ProcessMessageSeverity.Error,
                                    Description = @"Address Currency Code must be ""C""",
                                    Location = DeriveSbrElementLocation(violator, addOccurrenceIndex: true) + "/tns:AddressDetails.Currency.Code",
                                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.IEE.410139" } },
                                };


                                IDictionary<string, SBRElement> VRATOIEE410139_IEE153_Violator_TupleElementMap = violator.PopulatedChildElementsMap ?? new Dictionary<string, SBRElement>();

                                parameterOutput = "";

                                if (VRATOIEE410139_IEE153_Violator_TupleElementMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}AddressDetails.Currency.Code", out currentValue))
                                {
                                    if (!currentValue.IsNil) parameterOutput = currentValue.Value;
                                }

                                processMessage.Parameters.Add(new ProcessMessageParameter
                                { Name = "IEE153", Value = parameterOutput });

                                errors.Add(processMessage);
                            }
                        }
                    }
                }
            }

            #endregion

            #region VR.ATO.IEE.410139 - manual implementation for instance against IEE152

            /*  VR.ATO.IEE.410139
                        Address Currency Code must be 'C'

                    Legacy Rule Format:
                        IN TUPLE((iee.0003.lodge.request.xx.xx:TrusteeOrPartner) OR (iee.0003.lodge.request.xx.xx:FamilyTrust) OR (iee.0003.lodge.request.xx.xx:SpecifiedIndividual)) 
                        IN TUPLE(address2.xx.xx:AddressDetails) 
                        IF (pyde.xx.xx:AddressDetails.Currency.Code <> NULLORBLANK) AND (pyde.xx.xx:AddressDetails.Currency.Code <> "C")
                            RETURN VALIDATION MESSAGE 
                        ENDIF

                Technical Business Rule Format:
                        (^IEE130 <> BLANK) AND (^IEE130 <> 'C')

            Data Elements:

                        RP:^IEE130 = tns:AddressDetails.Currency.Code WHERE [AddressDetails.Usage.Code='POS'].[AddressDetails.Currency.Code='C']
                        */


            if (sbrSpecifiedIndividualCollection != null)
            {
                foreach (SBRElement sbrSpecifiedIndividual in sbrSpecifiedIndividualCollection)
                {

                    IDictionary<string, SBRElement> sbrSpecifiedIndividualTupleElementMap = sbrSpecifiedIndividual.PopulatedChildElementsMap ?? new Dictionary<string, SBRElement>();

                    IEnumerable<SBRElement> sbrSpecifiedIndividual_AddressDetailsCollection;
                    if (sbrSpecifiedIndividualTupleElementMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}AddressDetails", out currentValue))
                        sbrSpecifiedIndividual_AddressDetailsCollection = currentValue.Occurrences;
                    else
                        sbrSpecifiedIndividual_AddressDetailsCollection = null;
                    if (sbrRPContexts != null && sbrSpecifiedIndividual_AddressDetailsCollection != null && sbrSpecifiedIndividual_AddressDetailsCollection.Count() > 0)
                    {
                        IEnumerable<SBRElement> VRATOIEE410139_IEE152_CurrentParentTupleViolators = sbrSpecifiedIndividual_AddressDetailsCollection.Where(
                            ce => ce.ChildElements != null &&
                            ce.ChildElements.Any(te => te.Name == "AddressDetails.Currency.Code" && !string.IsNullOrWhiteSpace(te.Value) && te.Value != "C" &&
                            sbrRPContexts.Contains(te.Context)));
                        if (VRATOIEE410139_IEE152_CurrentParentTupleViolators != null && VRATOIEE410139_IEE152_CurrentParentTupleViolators.Count() > 0)
                        {
                            foreach (SBRElement violator in VRATOIEE410139_IEE152_CurrentParentTupleViolators)
                            {
                                processMessage = new ProcessMessageDocument()
                                {
                                    Code = "CMN.ATO.IEE.410139",
                                    Severity = ProcessMessageSeverity.Error,
                                    Description = @"Address Currency Code must be ""C""",
                                    Location = DeriveSbrElementLocation(violator, addOccurrenceIndex: false) + "/tns:AddressDetails.Currency.Code",
                                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.IEE.410139" } },
                                };


                                IDictionary<string, SBRElement> VRATOIEE410139_IEE152_Violator_TupleElementMap = violator.PopulatedChildElementsMap ?? new Dictionary<string, SBRElement>();

                                parameterOutput = "";

                                if (VRATOIEE410139_IEE152_Violator_TupleElementMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}AddressDetails.Currency.Code", out currentValue))
                                {
                                    if (!currentValue.IsNil) parameterOutput = currentValue.Value;
                                }

                                processMessage.Parameters.Add(new ProcessMessageParameter
                                { Name = "IEE152", Value = parameterOutput });

                                errors.Add(processMessage);
                            }
                        }
                    }
                }
            }
            #endregion
            #endregion

            #region VR.ATO.IEE.410265
            /*  VR.ATO.IEE.410265
                   Declaration Statement Type Code must be 'TrueAndCorrect'

               Legacy Rule Format:
                   IF [IEE147] <> NULLORBLANK AND
                   [IEE147] <> "TrueAndCorrect"
                     RETURN VALIDATION MESSAGE
                   ENDIF

                   [IEE147] = IEE:RP.TrueAndCorrect:pyin.02.03:Declaration.StatementType.Code IN TUPLE(iee.0003.lodge.req.02.00:Declaration)

               Technical Business Rule Format:
                   ^IEE147 <> BLANK AND ^IEE147 <> 'TrueAndCorrect'

               Data Elements:

                   RP:^IEE147 = tns:Declaration.StatementType.Code WHERE [Declaration.StatementType.Code = 'TrueAndCorrect']
            */
            IEnumerable<SBRElement> VRATOIEE410265_IEE147_Violators = null;
            if (sbrRPContexts != null && sbrDeclarationCollection != null && sbrDeclarationCollection.Count() > 0)
            {
                VRATOIEE410265_IEE147_Violators = sbrDeclarationCollection.Where(
                    ce => ce.ChildElements != null &&
                    ce.ChildElements.Any(te => te.Name == "Declaration.StatementType.Code" && te.Value != "TrueAndCorrect" &&
                    sbrRPContexts.Contains(te.Context)));
            }


            if (VRATOIEE410265_IEE147_Violators != null && VRATOIEE410265_IEE147_Violators.Count() > 0)
            {
                foreach (SBRElement violator in VRATOIEE410265_IEE147_Violators)
                {
                    processMessage = new ProcessMessageDocument()
                    {
                        Code = "CMN.ATO.IEE.410265",
                        Severity = ProcessMessageSeverity.Error,
                        Description = @"Declaration Statement Type Code must be 'TrueAndCorrect'",
                        Location = DeriveSbrElementLocation(violator, addOccurrenceIndex: true) + "/tns:Declaration.StatementType.Code",
                        Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.IEE.410265" } },
                    };


                    IDictionary<string, SBRElement> VRATOIEE410265_IEE147_Violator_TupleElementMap = violator.PopulatedChildElementsMap ?? new Dictionary<string, SBRElement>();

                    parameterOutput = "";
                    if (VRATOIEE410265_IEE147_Violator_TupleElementMap.TryGetValue("{http://www.sbr.gov.au/ato/iee}Declaration.StatementType.Code", out currentValue))
                    {
                        if (!currentValue.IsNil) parameterOutput = currentValue.Value;
                    }


                    processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "IEE147", Value = parameterOutput });

                    errors.Add(processMessage);
                }
            }

            #endregion
        }

        /// <summary>
        /// Derives the SBR element location.
        /// </summary>
        /// <param name="sbrElement">The SBR element.</param>
        /// <param name="addOccurrenceIndex">if set to <c>true</c> [add occurrence the occurrence index to the path].</param>
        /// <returns></returns>
        /// <remarks>Copied from IEE2018Consumer, since it is defined as private and hence inaccessible from a subclass</remarks>
        private static string DeriveSbrElementLocation(SBRElement sbrElement, bool addOccurrenceIndex = false)
        {
            if (sbrElement == null) return "/xbrli:xbrl";
            int xpathOccurrence = sbrElement.OccurrenceIndex + 1;
            string location;

            if (addOccurrenceIndex)
            {
                location = "/" + sbrElement.NamespacePrefix + ":" + sbrElement.Name + "[" +
                           xpathOccurrence + "]";
            }
            else
            {
                location = "/" + sbrElement.NamespacePrefix + ":" + sbrElement.Name;
            }

            if (sbrElement.IsInTuple)
            {
                location = DeriveSbrElementLocation(sbrElement.ParentTuple, true) + location;
            }
            else
            {
                location = "/xbrli:xbrl" + location;
            }

            return location;
        }
    }
}